#include "bsp.h"
#include <stdio.h>
#include <stdlib.h>

/* Procedure prototypes */
void spmd_start();
int allsums(int);
int sumI(int);
void printArray(int*,int);

/* Global variables */
char usage_str[] = "main: -p <number> -n <number> -v";
int N=1, nprocs=2, verbose=0;

/* ####################################################################
   ## The procedure "sumI" calculates the sum of p values stored in  ##
   ## p processors. That is given x_i on process i, then the result  ##
   ## of calling sumI(x_i) on process i will be x_0 + x+1 .. x_p-1   ##
   #################################################################### */

int sumI(int x) {
  int result;
  result = allsums(x);
  broadcast_ints(bsp_nprocs()-1,&result,&result,1);
  return result;
}

/* ####################################################################
   ## The procedure "allsums" calculates the running sums of p values##
   ## stored in p processors. That is given x_i on process i, then   ##
   ## the result of calling allsums(x_i) on process i will be        ##
   ## x_0 + x+1 .. x_i-1 + x_i.                                      ##
   #################################################################### */

int allsums(int x) {
  int i, left, right;
  bsp_push_reg(&left,sizeof(int));
  bsp_sync();

  right = x;
  for(i=1;i<bsp_nprocs();i*=2) {
    if (bsp_pid()+i < bsp_nprocs()) 
      bsp_put(bsp_pid()+i,&right,&left,0,sizeof(int));
    bsp_sync();
    if (bsp_pid()>=i) right = left + right;
  }
  bsp_pop_reg(&left);
  return right;
}

/* ####################################################################
   ## The procedure "broadcast_ints" transfers the array "src",      ##
   ## that contains nelem integers, held  on process "frompid" into  ##
   ## the array "dst" on all processes. "dst" has to contain at      ##
   ## least "nelem" integers.                                        ##
   #################################################################### */

void broadcast_ints(int frompid,int *src, int *dst, int nelem) {
  bsp_push_reg(src,nelem*sizeof(int));
  bsp_sync();

  bsp_get(frompid,src,0,dst,nelem*sizeof(int));
  bsp_sync();
  bsp_pop_reg(src);
}


/* ###################################################################
   ## "spmd_start" is a test harness which starts nprocs processes, ##
   ## and then does a total exchange of an array of "size", where   ##
   ## each element of the array contains the process ID.            ##
   ################################################################### */

void spmd_start() {
  int i,j,k;

  bsp_begin(nprocs);
   
    bsp_dtime();
    j = bsp_pid() + bsp_nprocs();
    k = sumI(j);
    for(i=0;i<bsp_nprocs();i++) {
      if (bsp_pid()==i) {
        printf("On process %d sumI(%d)=%d\n",i,j,k);
        fflush(stdout);
      }
      bsp_sync();
    }
  bsp_end();
}

/* ###################################################################
   ## The main procedure reads in the number of processes to spawn  ##
   ## off the command line. It then enters SPMD mode inside the     ##
   ## procedure spmd_start                                          ##
   ################################################################### */

void main(int argc,char **argv) {
  int i,temp, command_line_opts=1;
  double trash;

  bsp_init(spmd_start,argc,argv);
  i=1;
  while (command_line_opts) {
    if (i >= argc) {
      command_line_opts = 0;
    } else if ( (strcmp(argv[i],"-p")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      nprocs = temp;
      i += 2;
    } else if (strcmp(argv[i],"-v")==0) {
      verbose=1;
      i += 1;
    } else if ( (strcmp(argv[i],"-n")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      N = temp;
      i += 2;
    }  else {
      bsp_abort("{%s}: unknown option \"%s\"\n\t%s\n",
                argv[0],argv[i],usage_str);
    }
  }
  printf("Test of bsp_begin: only one process should print this\n");
    spmd_start();
  printf("Just about to finish, only process zero should "
         "get here (bsp_pid()=%d)\n",bsp_pid());
  exit(0);
}
