#include "bsp.h"
#include <stdio.h>
#include <stdlib.h>

/* Procedure prototypes */
void spmd_start();
int allsums(int);
void allsums3(int*,int);
int sumI(int);
void printArray(int*,int);

/* Global variables */
char usage_str[] = "main: -p <number> -n <number> -v";
int N=1, nprocs=2, verbose=0;
  
/* ####################################################################
   ## The procedure "allsums3" calculates the running sums of n      ##
   ##  values held in p processors; where n>p.                       ##
   #################################################################### */

void allsums3(int *array, int n_over_p) {
  int i, last, lefts_last;
  bsp_push_reg(&last,sizeof(int));

  for (i=1;i<n_over_p;i++)
    array[i] += array[i-1];

  last = allsums(array[n_over_p-1]);

  if (bsp_pid()==0) lefts_last=0;
  else 
    bsp_get(bsp_pid()-1,&last,0,&lefts_last,sizeof(int));
  bsp_sync();

  for(i=0;i<n_over_p;i++)
    array[i] += lefts_last;
  bsp_pop_reg(&last);
}

/* ####################################################################
   ## The procedure "allsums" calculates the running sums of p values##
   ## stored in p processors. That is given x_i on process i, then   ##
   ## the result of calling allsums(x_i) on process i will be        ##
   ## x_0 + x+1 .. x_i-1 + x_i.                                      ##
   #################################################################### */

int allsums(int x) {
  int i, left, right;
  bsp_push_reg(&left,sizeof(int));
  bsp_sync();

  right = x;
  for(i=1;i<bsp_nprocs();i*=2) {
    if (bsp_pid()+i < bsp_nprocs()) 
      bsp_put(bsp_pid()+i,&right,&left,0,sizeof(int));
    bsp_sync();
    if (bsp_pid()>=i) right = left + right;
  }
  bsp_pop_reg(&left);
  return right;
}


/* ###################################################################
   ## "spmd_start" is a test harness which starts nprocs processes, ##
   ## and then does a total exchange of an array of "size", where   ##
   ## each element of the array contains the process ID.            ##
   ################################################################### */

void spmd_start() {
  int i,j,k, *array;

  bsp_begin(nprocs);
   
    bsp_dtime();
    k = bsp_pid() + bsp_nprocs();
    array = calloc(k,sizeof(int));
    for(i=0;i<k;i++) array[i]=1+bsp_pid();
    allsums3(array,k);
    for(i=0;i<bsp_nprocs();i++) {
      if (bsp_pid()==i) {
        for(j=0;j<k;j++) 
          printf("On process %d allsums(%d)=%d\n",i,1+bsp_pid(),array[j]);
        fflush(stdout);
      }
      bsp_sync();
    }
  bsp_end();
}

/* ###################################################################
   ## The main procedure reads in the number of processes to spawn  ##
   ## off the command line. It then enters SPMD mode inside the     ##
   ## procedure spmd_start                                          ##
   ################################################################### */

void main(int argc,char **argv) {
  int i,temp, command_line_opts=1;
  double trash;

  bsp_init(spmd_start,argc,argv);
  i=1;
  while (command_line_opts) {
    if (i >= argc) {
      command_line_opts = 0;
    } else if ( (strcmp(argv[i],"-p")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      nprocs = temp;
      i += 2;
    } else if (strcmp(argv[i],"-v")==0) {
      verbose=1;
      i += 1;
    } else if ( (strcmp(argv[i],"-n")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      N = temp;
      i += 2;
    }  else {
      bsp_abort("{%s}: unknown option \"%s\"\n\t%s\n",
                argv[0],argv[i],usage_str);
    }
  }
  printf("Test of bsp_begin: only one process should print this\n");
    spmd_start();
  printf("Just about to finish, only process zero should "
         "get here (bsp_pid()=%d)\n",bsp_pid());
  exit(0);
}
