/* ###################################################################
   ## E x e r c i s e  T h r e e : model answer                     ##
   ################################################################### */
#include "bsp.h"
#include <stdio.h>
#include <stdlib.h>

/* Procedure prototypes */
void spmd_start();
void broadcast_ints(int,int*,int*,int);
void printArray(int*,int);

/* Global variables */
char usage_str[] = "main: -p <number> -n <number> -v";
int N=1, nprocs=2, verbose=0;
  

/* ####################################################################
   ## The procedure "broadcast_ints" transfers the array "src",      ##
   ## that contains nelem integers, held  on process "frompid" into  ##
   ## the array "dst" on all processes. "dst" has to contain at      ##
   ## least "nelem" integers.                                        ##
   #################################################################### */

void broadcast_ints(int frompid,int *src, int *dst, int nelem) {
  int i;
  bsp_push_reg(dst,nelem*sizeof(int));
  bsp_sync();

  if (bsp_pid()==frompid)
    for(i=0;i<bsp_nprocs();i++)
      bsp_put(i,src,dst,0,nelem*sizeof(int));
  bsp_sync();
  bsp_pop_reg(dst);
}


/* ###################################################################
   ## "spmd_start" is a test harness which starts nprocs processes, ##
   ## and then does a total exchange of an array of "size", where   ##
   ## each element of the array contains the process ID.            ##
   ################################################################### */

void spmd_start() {
  int i,j,*array;

  bsp_begin(nprocs);
  /* Broadcast the value of N stored on process zero */
    broadcast_ints(0,&N,&N,1);
    broadcast_ints(0,&verbose,&verbose,1);

    printf("On process %d, p=%d n=%d\n",bsp_pid(),bsp_nprocs(),N);

    array = calloc(N,sizeof(int));
    if (array==NULL)
      bsp_abort("Unable to allocate %d element array",N);
    for(i=0;i<N;i++) 
      array[i] = bsp_pid() + i;
    
    bsp_dtime();
    broadcast_ints(bsp_nprocs()-1,array,array,N);
    if (bsp_pid()==0) printf("Broadcast N ints = %f secs\n",bsp_dtime());
    if (verbose) printArray(array,N); 
  bsp_end();
}


/* ###################################################################
   ## The main procedure reads in the number of processes to spawn  ##
   ## off the command line. It then enters SPMD mode inside the     ##
   ## procedure spmd_start                                          ##
   ################################################################### */

void main(int argc,char **argv) {
  int i,temp, command_line_opts=1;
  double trash;

  bsp_init(spmd_start,argc,argv);
  i=1;
  while (command_line_opts) {
    if (i >= argc) {
      command_line_opts = 0;
    } else if ( (strcmp(argv[i],"-p")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      nprocs = temp;
      i += 2;
    } else if (strcmp(argv[i],"-v")==0) {
      verbose=1;
      i += 1;
    } else if ( (strcmp(argv[i],"-n")==0) &&
                (sscanf(argv[i+1],"%d",&temp))) {
      N = temp;
      i += 2;
    }  else {
      bsp_abort("{%s}: unknown option \"%s\"\n\t%s\n",
                argv[0],argv[i],usage_str);
    }
  }
  printf("Test of bsp_begin: only one process should print this\n");
    spmd_start();
  printf("Just about to finish, only process zero should "
         "get here (bsp_pid()=%d)\n",bsp_pid());
  exit(0);
}


/* ###################################################################
   ## The procedure "printArray" serialises the printing of the     ##
   ## array "array" of size "size" held on each process.            ##
   ################################################################### */

void printArray(int *array,int size) {
  int i,j;

  for(i=0;i<bsp_nprocs();i++) {
    if (bsp_pid()==i) {
      for(j=0;j<size;j++) {
        printf("on %d array[%d]=%d\n",bsp_pid(),j,array[j]);
      }
      fflush(stdout);
    }
    bsp_sync();
  }
}
